{
Volume3D
3D function library for CAM purposes

program author
Arndt Kritzner, see http://www.kritzner.de

version
1.0.1 ... 2003-01-18
1.0 ... 2002-09-07

licenses:
volume3d is available under the terms of GNUs GPL (for developers of free software,
only FreePascal or GNU Pascal with Linux) or alternatively under a commercial
license (for developers of commercial software)

All rights reserved.
Any usage in another kind than described in the purchased license is prohibited.
If you want to choose the GNU GPL on another platform than GNU/Linux ask for a
free license for your preferred OS mailto:arndt@kritzner.de.

See http://www.kritzner.de/pages/v3d/v3d.html for technical and licensing details.
}

{
Unit export3d
file import/export (stl und gts file format) functions for 3D models
}

unit export3d;

interface

uses math,objects,triangle3d;

function export_gts(o:pshape_3d;filename:string):boolean;
function import_gts(filename:string):pshape_3d;

function export_stl(o:pshape_3d;filename:string):boolean;
function import_stl(filename:string;max_:tri_idx3d):pshape_3d;

implementation

type tcompare = (gt,eq,lt);

		 plist_point=^tlist_point;
		 tlist_point=record
									 next:plist_point;
									 p:tpoint;
								 end;

		 tgts_edge=record
    					   p1,p2:tri_idx3d;
    					 end;		

		 plist_edge=^tlist_edge;	
		 tlist_edge=record
									next:plist_edge;
									p1,p2:tri_idx3d;
								end;

		 tgts_triangle=record
									   l1,l2,l3:tri_idx3d;
									 end;	

		 plist_triangle=^tlist_triangle;
		 tlist_triangle=record
											next:plist_triangle;
											l1,l2,l3:tri_idx3d;
										end;

var plist:plist_point;
		elist:plist_edge;
		tlist:plist_triangle;
			
function geomsort(p,p_comp:tpoint):tcompare;
{$ifndef FKT_RESULT}
var result:tcompare;
{$endif}
begin
  result:=eq;
	if not same_point(p,p_comp) then begin
    if result=eq then
      if p.x>p_comp.x+eps then
        result:=gt
      else
        if p.x<p_comp.x-eps then
          result:=lt;
    if result=eq then
      if p.y>p_comp.y+eps then
        result:=gt
      else
        if p.y<p_comp.y-eps then
          result:=lt;
    if result=eq then
      if p.z>p_comp.z+eps then
        result:=gt
      else
        if p.z<p_comp.z-eps then
          result:=lt;
  end;
{$ifndef FKT_RESULT}
  geomsort:=result;
{$endif}
end;

function point_idx(p:tpoint):tri_idx3d;
var element,newelement:plist_point;
		isnew,inserted:boolean;
		compval:tcompare;
{$ifndef FKT_RESULT}
		result:tri_idx3d;
{$endif}
begin
	result:=1;
	isnew:=false;
	if assigned(plist) then begin
    inserted:=false;
    element:=plist;
		while not inserted do begin
			compval:=geomsort(p,element^.p);
			if compval=gt then begin
				inc(result);
        if assigned(element^.next) then
          element:=element^.next
				else begin
					new(element^.next);
					element:=element^.next;
					element^.next:=nil;
					inserted:=true;
					isnew:=true;
				end;
			end else
				if compval=lt then begin
					new(newelement);
					newelement^.next:=element^.next;
					element^.next:=newelement;
					newelement^.p:=element^.p;
          inserted:=true;
          isnew:=true;
        end else
          inserted:=true;
		end;
	end else begin
		new(element);
		element^.next:=nil;
		plist:=element;
		isnew:=true;
	end;
	if isnew then
		element^.p:=p;
{$ifndef FKT_RESULT}
	point_idx:=result;
{$endif}
end;

function point_append(p:tpoint):tri_idx3d;
var element:plist_point;
{$ifndef FKT_RESULT}
		result:tri_idx3d;
{$endif}
begin
	result:=1;
	if assigned(plist) then begin
    element:=plist;
		while assigned(element^.next) do begin
			inc(result);
      element:=element^.next;
    end;
    inc(result);
		new(element^.next);
		element:=element^.next;
		element^.next:=nil;
	end else begin
		new(element);
		element^.next:=nil;
		plist:=element;
	end;
	element^.p:=p;
{$ifndef FKT_RESULT}
	point_append:=result;
{$endif}
end;

function idx_point(idx:tri_idx3d):tpoint;
var i:tri_idx3d;	
		element:plist_point;
begin
	i:=1;
  element:=plist;
	while i<idx do begin
		element:=element^.next;
		inc(i);
	end;
	idx_point:=element^.p;
end;

function count_points:tri_idx3d;
var element:plist_point;
{$ifndef FKT_RESULT}
		result:tri_idx3d;
{$endif}
begin
  result:=0;
  element:=plist;
	while assigned(element) do begin
		element:=element^.next;
		inc(result);
	end;
{$ifndef FKT_RESULT}
	count_points:=result;
{$endif}
end;

procedure free_points(var list:plist_point);
var remark:plist_point;
begin
  while assigned(list) do begin
		remark:=list^.next;
    dispose(list);
		list:=remark;
  end;
  list:=nil;
end;

function edge_idx(p1,p2:tri_idx3d):tri_idx3d;
var element,newelement:plist_edge;
		isnew,inserted:boolean;
		compval:tcompare;
{$ifndef FKT_RESULT}
		result:tri_idx3d;
{$endif}
begin
	if p1>p2 then
    swap(p1,p2);
	result:=1;
	isnew:=false;
	if assigned(elist) then begin
    inserted:=false;
    element:=elist;
		while not inserted do begin
			if (p1=element^.p1) and (p2=element^.p2) then
        compval:=eq
			else
				if (p1<element^.p1) or ((p1=element^.p1) and (p2<element^.p2)) then
          compval:=lt
				else
          compval:=gt;
			if compval=gt then begin
				inc(result);
        if assigned(element^.next) then
          element:=element^.next
				else begin
					new(element^.next);
					element:=element^.next;
					element^.next:=nil;
					inserted:=true;
					isnew:=true;
				end;
			end else
				if compval=lt then begin
					new(newelement);
					newelement^.next:=element^.next;
					element^.next:=newelement;
					newelement^.p1:=element^.p1;
					newelement^.p2:=element^.p2;
          inserted:=true;
          isnew:=true;
        end else
          inserted:=true;
		end;
	end else begin
		new(element);
		element^.next:=nil;
		elist:=element;
		isnew:=true;
	end;
	if isnew then begin
		element^.p1:=p1;
		element^.p2:=p2;
  end;
{$ifndef FKT_RESULT}
	edge_idx:=result;
{$endif}
end;

function edge_append(p1,p2:tri_idx3d):tri_idx3d;
var element:plist_edge;
{$ifndef FKT_RESULT}
		result:tri_idx3d;
{$endif}
begin
	result:=1;
	if assigned(elist) then begin
    element:=elist;
		while assigned(element^.next) do begin
			inc(result);
      element:=element^.next;
    end;
    inc(result);
		new(element^.next);
		element:=element^.next;
		element^.next:=nil;
	end else begin
		new(element);
		element^.next:=nil;
		elist:=element;
	end;
	element^.p1:=p1;
	element^.p2:=p2;
{$ifndef FKT_RESULT}
	edge_append:=result;
{$endif}
end;

function idx_edge(idx:tri_idx3d):tgts_edge;
var i:tri_idx3d;	
		element:plist_edge;
{$ifndef FKT_RESULT}
		result:tgts_edge;
{$endif}
begin
	i:=1;
  element:=elist;
	while i<idx do begin
		element:=element^.next;
		inc(i);
	end;
	result.p1:=element^.p1;
	result.p2:=element^.p2;
{$ifndef FKT_RESULT}
	idx_edge:=result;
{$endif}
end;

function count_edges:tri_idx3d;
var element:plist_edge;
{$ifndef FKT_RESULT}
		result:tri_idx3d;
{$endif}
begin
  result:=0;
  element:=elist;
	while assigned(element) do begin
		element:=element^.next;
		inc(result);
	end;
{$ifndef FKT_RESULT}
	count_edges:=result;
{$endif}
end;

procedure free_edges(var list:plist_edge);
var remark:plist_edge;
begin
  while assigned(list) do begin
		remark:=list^.next;
    dispose(list);
		list:=remark;
  end;
  list:=nil;
end;

function sort_triangles(l1,l2,l3:tri_idx3d):tgts_triangle;
var swapped:boolean;
{$ifndef FKT_RESULT}
		result:tgts_triangle;
{$endif}
begin
	result.l1:=l1;
	result.l2:=l2;
	result.l3:=l3;
	swapped:=true;
  while swapped do begin
    swapped:=false;
		if result.l1>result.l2 then begin
      swap(result.l1,result.l2);
			swapped:=true;
    end;
		if result.l2>result.l3 then begin
      swap(result.l2,result.l3);
			swapped:=true;
    end;
  end;
{$ifndef FKT_RESULT}
	sort_triangles:=result;
{$endif}
end;

function geomsort(t:tgts_triangle;l1,l2,l3:tri_idx3d):tcompare;
{$ifndef FKT_RESULT}
var result:tcompare;
{$endif}
begin
  result:=eq;
	if (t.l1>l1) then
    result:=gt;
	if (t.l1<l1) then
    result:=lt;
  if result=eq then begin
  	if (t.l2>l2) then
      result:=gt;
  	if (t.l2<l2) then
      result:=lt;
  end;
  if result=eq then begin
  	if (t.l3>l3) then
      result:=gt;
  	if (t.l3<l3) then
      result:=lt;
  end;
{$ifndef FKT_RESULT}
	geomsort:=result;
{$endif}
end;

function triangle_idx(l1,l2,l3:tri_idx3d):tri_idx3d;
var element,newelement:plist_triangle;
		isnew,inserted:boolean;
		compval:tcompare;
		tri_edges:tgts_triangle;
{$ifndef FKT_RESULT}
		result:tri_idx3d;
{$endif}
begin
	tri_edges:=sort_triangles(l1,l2,l3);
	result:=1;
	isnew:=false;
	if assigned(tlist) then begin
    inserted:=false;
    element:=tlist;
		while not inserted do begin
			compval:=geomsort(tri_edges,element^.l1,element^.l2,element^.l3);
			if compval=gt then begin
				inc(result);
        if assigned(element^.next) then
          element:=element^.next
				else begin
					new(element^.next);
					element:=element^.next;
					element^.next:=nil;
					inserted:=true;
					isnew:=true;
				end;
			end else
				if compval=lt then begin
					new(newelement);
					newelement^.next:=element^.next;
					element^.next:=newelement;
					newelement^.l1:=element^.l1;
					newelement^.l2:=element^.l2;
					newelement^.l3:=element^.l3;
          inserted:=true;
          isnew:=true;
        end else
          inserted:=true;
		end;
	end else begin
		new(element);
		element^.next:=nil;
		tlist:=element;
		isnew:=true;
	end;
	if isnew then begin
		element^.l1:=tri_edges.l1;
		element^.l2:=tri_edges.l2;
		element^.l3:=tri_edges.l3;
  end;
{$ifndef FKT_RESULT}
	triangle_idx:=result;
{$endif}
end;

function idx_triangle(idx:tri_idx3d):tgts_triangle;
var i:tri_idx3d;	
		element:plist_triangle;
{$ifndef FKT_RESULT}
		result:tgts_triangle;		
{$endif}
begin
	i:=1;
  element:=tlist;
	while i<idx do begin
		element:=element^.next;
		inc(i);
	end;
	result.l1:=element^.l1;
	result.l2:=element^.l2;
	result.l3:=element^.l3;
{$ifndef FKT_RESULT}
	idx_triangle:=result;
{$endif}
end;

function count_triangles:tri_idx3d;
var element:plist_triangle;
{$ifndef FKT_RESULT}
		result:tri_idx3d;
{$endif}
begin
  result:=0;
  element:=tlist;
	while assigned(element) do begin
		element:=element^.next;
		inc(result);
	end;
	count_triangles:=result;
end;

procedure free_triangles(var list:plist_triangle);
var remark:plist_triangle;
begin
  while assigned(list) do begin
		remark:=list^.next;
    dispose(list);
		list:=remark;
  end;
  list:=nil;
end;

function export_gts(o:pshape_3d;filename:string):boolean;
var idx:tri_idx3d;
		gts_file:text;
		t:ptriangle;
		p:array[1..3] of tri_idx3d;
		e:array[1..3] of tri_idx3d;
		cp,ce,ct:tri_idx3d;
{$ifndef FKT_RESULT}
		result:boolean;
{$endif}
begin
	result:=false;
  if o^.count>0 then begin
		Assign(gts_file,filename);
		{$i-}
		Rewrite(gts_file);
		{$i+}
		if IOResult=0 then begin
  	  for idx:=0 to o^.count-1 do begin
		  	t:=o^.get_triangle(idx);
				point_idx(t^.p1);
				point_idx(t^.p2);
				point_idx(t^.p3);
  		end;
  	  for idx:=0 to o^.count-1 do begin
		  	t:=o^.get_triangle(idx);
				p[1]:=point_idx(t^.p1);
				p[2]:=point_idx(t^.p2);
				p[3]:=point_idx(t^.p3);
				edge_idx(p[1],p[2]);
				edge_idx(p[1],p[3]);
				edge_idx(p[2],p[3]);
  		end;
  	  for idx:=0 to o^.count-1 do begin
		  	t:=o^.get_triangle(idx);
				p[1]:=point_idx(t^.p1);
				p[2]:=point_idx(t^.p2);
				p[3]:=point_idx(t^.p3);
				e[1]:=edge_idx(p[1],p[2]);
				e[2]:=edge_idx(p[1],p[3]);
				e[3]:=edge_idx(p[2],p[3]);
				triangle_idx(e[1],e[2],e[3]);
  		end;
			cp:=count_points;
			ce:=count_edges;
			ct:=count_triangles;
			if (cp>0) and (ce>0) and (ct>0) then begin
  			WriteLn(gts_file,cp,' ',ce,' ',ct);
  			for idx:=1 to cp do
  				with idx_point(idx) do
            WriteLn(gts_file,x,' ',y,' ',z);
  			for idx:=1 to ce do
  				with idx_edge(idx) do
            WriteLn(gts_file,p1,' ',p2);
  			for idx:=1 to ct do
  				with idx_triangle(idx) do
            WriteLn(gts_file,l1,' ',l2,' ',l3);
				result:=true;
			end;
			Close(gts_file);
  		free_triangles(tlist);
  		free_edges(elist);
  		free_points(plist);
		end;
  end;
{$ifndef FKT_RESULT}
	export_gts:=result;
{$endif}
end;

function import_gts(filename:string):pshape_3d;
var idx,points,edges,triangles,p1,p2,e1,e2,e3:tri_idx3d;
		p_idx:array[1..3] of tri_idx3d;
		gts_file:text;
		p_count:byte;
		error:boolean;
		p:tpoint;
		t:ptriangle;
{$ifndef FKT_RESULT}
		result:pshape_3d;
{$endif}
function new_point(p:tri_idx3d):boolean;
var idx:byte;
{$ifndef FKT_RESULT}
		result:boolean;
{$endif}
begin
  result:=true;
	for idx:=1 to p_count do
	  if p=p_idx[idx] then
	    result:=false;
	if result then begin
		inc(p_count);
		p_idx[p_count]:=p;
	end;
{$ifndef FKT_RESULT}
	new_point:=result;
{$endif}
end;
begin
	result:=Nil;
	error:=True;
	Assign(gts_file,filename);
	{$i-}
	Reset(gts_file);
	{$i+}
	if IOResult=0 then begin
		error:=false;
		{$i-}
		ReadLN(gts_file,points,edges,triangles);
		{$i+}
		if IOResult<>0 then begin
		  error:=True;
		  points:=0;
		  edges:=0;
		  triangles:=0;
		end;
		if (points>0) and (triangles>0) then begin
  		for idx:=0 to points-1 do begin
  			ReadLN(gts_file,p.x,p.y,p.z);
  			point_append(p);
  		end;
  		for idx:=0 to edges-1 do begin
  		  ReadLN(gts_file,p1,p2);
  		  edge_append(p1,p2);
  		end;
  		if not error then
  			New(result,init(triangles,'GTS'));
  		for idx:=0 to triangles-1 do begin
	  		p_count:=0;
  			ReadLN(gts_file,e1,e2,e3);
  			new_point(idx_edge(e1).p1);
  			new_point(idx_edge(e1).p2);
  			new_point(idx_edge(e2).p1);
  			new_point(idx_edge(e2).p2);
  			new_point(idx_edge(e3).p1);
  			new_point(idx_edge(e3).p2);
  			if p_count=3 then begin
  				new(t);
    			t^.p1:=idx_point(p_idx[1]);
    			t^.p2:=idx_point(p_idx[2]);
    			t^.p3:=idx_point(p_idx[3]);
    			result^.ins_triangle(t);
  			end;
  		end;
  		free_triangles(tlist);
  		free_points(plist);
		end;
	  Close(gts_file);
	end;
{$ifndef FKT_RESULT}
	import_gts:=result;
{$endif}
end;

function export_stl(o:pshape_3d;filename:string):boolean;
var idx:tri_idx3d;
		stl_file:text;
		t:ptriangle;
		n:tpoint;
{$ifndef FKT_RESULT}
		result:boolean;
{$endif}
begin
	result:=false;
  if o^.count>0 then begin
		Assign(stl_file,filename);
		{$i-}
		Rewrite(stl_file);
		{$i+}
		if IOResult=0 then begin
      WriteLn(stl_file,'solid');
  	  for idx:=0 to o^.count-1 do begin
		  	t:=o^.get_triangle(idx);
		  	with tri_norm(t) do begin
		  	  n:=norm_vect(point_point(p1,p2));
		  	end;
		  	if vec_len(n)>1-eps then begin
  				Writeln(stl_file,'facet normal ',n.x,' ',n.y,' ',n.z);
          WriteLn(stl_file,'  outer loop');
          WriteLn(stl_file,'    vertex ',t^.p1.x,' ',t^.p1.y,' ',t^.p1.z);
          WriteLn(stl_file,'    vertex ',t^.p2.x,' ',t^.p2.y,' ',t^.p2.z);
          WriteLn(stl_file,'    vertex ',t^.p3.x,' ',t^.p3.y,' ',t^.p3.z);
          WriteLn(stl_file,'  endloop');
          WriteLn(stl_file,'endfacet');
        end;
  		end;
      WriteLn(stl_file,'endsolid');
			Close(stl_file);
			result:=true;
		end;
	end;
{$ifndef FKT_RESULT}
	export_stl:=result;
{$endif}
end;

function line_2_point(line:string):tpoint;
var	value:string;
		idx:byte;
		error:integer;
{$ifndef FKT_RESULT}
		result:tpoint;
{$endif}
begin
	result.x:=0;
	result.y:=0;
	result.z:=0;
	idx:=0;
	while length(line)>0 do begin
	  if pos(' ',line)=1 then
	    delete(line,1,1)
	  else if pos('vertex',line)=1 then
	    delete(line,1,length('vertex'))
	  else begin
	  	inc(idx);
	    if pos(' ',line)>0 then begin
	      value:=copy(line,1,pred(pos(' ',line)));
	      delete(line,1,length(value));
	    end else begin
				value:=line;
				line:='';
	    end;
	    case idx of
	    	1:val(value,result.x,error);
	    	2:val(value,result.y,error);
	    	3:val(value,result.z,error);
	    end;
	  end;
	end;
{$ifndef FKT_RESULT}
	line_2_point:=result;
{$endif}
end;

function import_stl(filename:string;max_:tri_idx3d):pshape_3d;
var stl_file:text;
		readstate:array[1..2] of byte;
		error:boolean;
		p:array[1..3] of tpoint;
		t:ptriangle;
		line:string;
{$ifndef FKT_RESULT}
		result:pshape_3d;
{$endif}
begin
	result:=Nil;
	error:=True;
	readstate[1]:=0;
	readstate[2]:=0;
	Assign(stl_file,filename);
	{$i-}
	Reset(stl_file);
	{$i+}
	if IOResult=0 then begin
		error:=false;
		New(result,init(max_,'GTS'));
		while not (EOF(stl_file) or error) do begin
  		{$i-}
  		ReadLN(stl_file,line);
  		{$i+}
  		if IOResult<>0 then
  		  error:=True
  		else begin
  			if pos('solid',line)=1 then begin
  			  readstate[1]:=1;readstate[2]:=0;
  			end else if pos('endsolid',line)=1 then begin
  			  readstate[1]:=2;readstate[2]:=0;
  			end else
  			  if readstate[1]=1 then begin
  			    if pos('facet normal',line)>0 then
  			    	readstate[2]:=0
  			    else if pos('endfacet',line)>0 then
  			    	readstate[2]:=0
  			    else if pos('outer loop',line)>0 then
  			    	readstate[2]:=0
  			    else if pos('endloop',line)>0 then
  			    	readstate[2]:=0
  			    else if pos('vertex',line)>0 then
  			      inc(readstate[2]);
  			    if readstate[2]>0 then begin
  			    	p[readstate[2]]:=line_2_point(line);
  			    	if readstate[2]=3 then begin
          			new(t);
          			t^.p1:=p[1];
          			t^.p2:=p[2];
          			t^.p3:=p[3];
          			result^.ins_triangle(t);
        			end;
      			end;
  			  end;
  		end;
		end;
	  Close(stl_file);
	end;
{$ifndef FKT_RESULT}
	import_stl:=result;
{$endif}
end;

begin
	plist:=nil;
	elist:=nil;
	tlist:=nil;
end.

