{
Volume3D
3D function library for CAM purposes

program author
Arndt Kritzner, see http://www.kritzner.de

version
1.0.1 ... 2003-01-18
1.0 ... 2002-09-07

licenses:
volume3d is available under the terms of GNUs GPL (for developers of free software,
only FreePascal or GNU Pascal with Linux) or alternatively under a commercial
license (for developers of commercial software)

All rights reserved.
Any usage in another kind than described in the purchased license is prohibited.
If you want to choose the GNU GPL on another platform than GNU/Linux ask for a
free license for your preferred OS mailto:arndt@kritzner.de.

See http://www.kritzner.de/pages/v3d/v3d.html for technical and licensing details.
}

{
Unit trans3d
transformations (translation, rotation and scaling) for 3D models
}

unit trans3d;

interface

uses math,objects,triangle3d;

function translate(x,y,z:float3d; o:pshape_3d):boolean;
function rotate(x1,y1,z1,x2,y2,z2,a:float3d; o:pshape_3d):boolean;
function scale(x,y,z,s:float3d; o:pshape_3d):boolean;

implementation

function translate(x,y,z:float3d; o:pshape_3d):boolean;
{Verschiebung einer Form um x,y,z}
var t:ptriangle;
		idx:tri_idx3d;
{$ifndef FKT_RESULT}		
		result:boolean;
{$endif}
begin
	result:=false;
	if o^.count > 0 then begin
	  for idx:=0 to o^.count-1 do begin
	  	t:=o^.get_triangle(idx);
	  	t^.p1:=vec_add(t^.p1,point(x,y,z));
	  	t^.p2:=vec_add(t^.p2,point(x,y,z));
	  	t^.p3:=vec_add(t^.p3,point(x,y,z));
		end;
		result:=true;
	end;
{$ifndef FKT_RESULT}		
	translate:=result;
{$endif}
end;

function rotate(x1,y1,z1,x2,y2,z2,a:float3d; o:pshape_3d):boolean;
{Rotation einer Form um die Achse (x1,y1,z1)->(x2,y2,z2) mit dem Winkel a}
var t:ptriangle;
		l:tline;
		pa,p,p90,pn:tpoint;
		idx:tri_idx3d;
		pidx:byte;
{$ifdef LOG}
		A_old,A_new:float3d;
{$endif}
{$ifndef FKT_RESULT}		
		result:boolean;
{$endif}
begin
	result:=false;
	l.p1:=point(x1,y1,z1);
	l.p2:=point(x2,y2,z2);
	if (o^.count>0) and (vec_len(point_point(l.p1,l.p2))>0) then begin
	  for idx:=0 to o^.count-1 do begin
	  	t:=o^.get_triangle(idx);
{$ifdef LOG}
			A_old:=tri_area(t);
{$endif}
			for pidx:=1 to 3 do begin
  			case pidx of
					1:pa:=t^.p1;
					2:pa:=t^.p2;
					3:pa:=t^.p3;
				end;
  			p:=perp_point_line(pa,l);
{$ifdef LOG}
  			if TriLog then begin		
  				writeln('Punkt: (',pa.x:10:2,pa.y:10:2,pa.z:10:2,'), Lotpunkt: (',p.x:10:2,p.y:10:2,p.z:10:2,')');
  				writeln('Linie: (',l.p1.x:10:2,l.p1.y:10:2,l.p1.z:10:2,')->(',l.p2.x:10:2,l.p2.y:10:2,l.p2.z:10:2,')');
  			end;
{$endif}
  			if vec_len(point_point(pa,p))>0 then begin
  				p90:=vec_X_vec(point_point(l.p1,l.p2),point_point(p,pa));
{$ifdef LOG}
  				if TriLog then
  					writeln('90-Grad-Vektor: (',p90.x:10:2,p90.y:10:2,p90.z:10:2,')');
{$endif}
  				p90:=vec_scale(p90,vec_len(point_point(p,pa))/vec_len(p90));
					pn:=vec_add(p,vec_add(vec_scale(point_point(p,pa),cos(a)),vec_scale(p90,sin(a))));
{$ifdef LOG}
					if TriLog then
						writeln('90-Grad-Vektor: (',p90.x:10:2,p90.y:10:2,p90.z:10:2,'), gedrehter Punkt: (',pn.x:10:2,pn.y:10:2,pn.z:10:2,')');
{$endif}
  			end else
					pn:=pa;
				case pidx of
					1:t^.p1:=pn;
					2:t^.p2:=pn;
					3:t^.p3:=pn;
				end;
			end;
{$ifdef LOG}
  		if TriLog then begin		
  			A_new:=tri_area(t);
  			if A_new<>A_old then
  				writeln('Flaecheninhalt verfaelscht: ',A_old:10:2,' ==> ',A_new:10:2);
  		end;
{$endif}
		end;
		result:=true;
	end;
{$ifndef FKT_RESULT}		
	rotate:=result;
{$endif}
end;

function scale(x,y,z,s:float3d; o:pshape_3d):boolean;
{Streckung einer Form um das Zentrum (x,y,z) mit dem Faktor s}
var t:ptriangle;
		idx:tri_idx3d;
{$ifndef FKT_RESULT}		
		result:boolean;
{$endif}
begin
	result:=false;
	if o^.count > 0 then begin
	  for idx:=0 to o^.count-1 do begin
	  	t:=o^.get_triangle(idx);
	  	t^.p1:=vec_add(point(x,y,z),vec_scale(vec_add(t^.p1,point(-x,-y,-z)),s));
	  	t^.p2:=vec_add(point(x,y,z),vec_scale(vec_add(t^.p2,point(-x,-y,-z)),s));
	  	t^.p3:=vec_add(point(x,y,z),vec_scale(vec_add(t^.p3,point(-x,-y,-z)),s));
		end;
		result:=true;
	end;
{$ifndef FKT_RESULT}		
	scale:=result;
{$endif}
end;

end.

