{
Volume3D
3D function library for CAM purposes

program author
Arndt Kritzner, see http://www.kritzner.de

version
1.0.1 ... 2003-01-18
1.0 ... 2002-09-07

licenses:
volume3d is available under the terms of GNUs GPL (for developers of free software,
only FreePascal or GNU Pascal with Linux) or alternatively under a commercial
license (for developers of commercial software)

All rights reserved.
Any usage in another kind than described in the purchased license is prohibited.
If you want to choose the GNU GPL on another platform than GNU/Linux ask for a
free license for your preferred OS mailto:arndt@kritzner.de.

See http://www.kritzner.de/pages/v3d/v3d.html for technical and licensing details.
}

{
Unit triangle3d
basic operations and types for triangulated shape surfaces
}

unit triangle3d;

interface

uses objects;		

const tf_init=0;
			tf_mark=1;
			tf_delete=128;
type	float3d=double;
			poly_idx3d=word;
			tri_idx3d=cardinal;

			tsort=(none_,X_,Y_,Z_);

			tpoint=record
							x,y,z:float3d;
						end;
						
      ppoly=^tpoly;
		  tpoly=array[poly_idx3d] of tpoint;

			pline=^tline;
			tline=record
							p1,p2:tpoint;
						end;

			ptriangle=^ttriangle;
			ttriangle=record
									flag:byte;
									p1,p2,p3:tpoint;
								end;

			ttri_tri_rel=(tt_none,tt_all,tt_back,tt_surf,tt_pane,tt_surf_pane,tt_back_pane,tt_back_2_back,tt_back_2_surf,tt_surf_2_back,tt_surf_2_surf,tt_pane_double,tt_pane_anti);
			stri_tri_rel=set of ttri_tri_rel;
			io_rel=(io_in,io_border,io_out,io_split);
			sio_rel=set of io_rel;
								
			ttrianglerelation=record
													d:float3d;
													r:ttri_tri_rel;
												end;	

			tpane=record
							x,y,z,c:double;
			      end;

			ptri_tri_line=^ttri_tri_line;
			ttri_tri_line=record
											p1,p2:tpoint;
											ok:boolean;
										end;

			pX_point=^tX_point;
			tX_point=record
								p:tpoint;
								in_pane:boolean;
								ok:boolean;
							end;
							
			tsplit_tri=record
									 t:array[1..3] of ptriangle;
									 l:pline;
									 ok:boolean;
								 end;
								
			ttrans_rs=record
									axis:tpoint;
									angle:float3d;
									scale:float3d;
								end;

			ptrianglearray=^ttrianglearray;
			ttrianglearray=array[0..$ffffff] of ptriangle;

			size_mode=(cl_minimize,cl_exact,cl_maximize);
			
			pshape_3d=^tshape_3d;
			tshape_3d=object(tobject)
  							private
                  tr:ptrianglearray;
                  idx_max:tri_idx3d;
                  sort:tsort;
 								public
                  name:string;
                  count:tri_idx3d;
                  constructor init(size:tri_idx3d);
                  constructor init(size:tri_idx3d;nam:string);
                  function ins_triangle(idx:tri_idx3d;t:ptriangle):boolean;
                  function ins_triangle(t:ptriangle):boolean;
                  function del_triangle(idx:tri_idx3d):boolean;
                  procedure del_triangles(x1,y1,z1,x2,y2,z2:float3d;rel:sio_rel);
                  function get_triangle(idx:tri_idx3d):ptriangle;
                  procedure list_triangles(x1,y1,z1,x2,y2,z2:float3d;rel:sio_rel);
                  procedure list_triangles(x,y,z:float3d;rel:sio_rel);
                  procedure list_triangles;
									procedure inside2out(mask:byte);
									procedure inside2out;
									procedure change_size(sm:size_mode);
									function clone_shape(nam:string;sm:size_mode):pshape_3d;
									function clone_shape(nam:string):pshape_3d;
									procedure dup_triangle(i:tri_idx3d;var t:ttriangle);
									function include_triangles(s:pshape_3d):boolean;
									function break_triangles(t:ptriangle):boolean;
									function sum_area:float3d;
                  destructor done; virtual;
  							end;

function min(v1,v2:float3d):float3d;  							
function max(v1,v2:float3d):float3d;  							
procedure swap(var a,b:float3d);
procedure swap(var a,b:tri_idx3d);
function point(x,y,z:float3d):tpoint;
function same_point(p1,p2:tpoint):boolean;
function point_point(p1,p2:tpoint):tpoint;
function vec_len(v:tpoint):float3d;
function vec_add(p1,p2:tpoint):tpoint;
function vec_scale(v:tpoint;s:float3d):tpoint;
function vec_p_vec(v1,v2:tpoint):float3d;
function vec_X_vec(v1,v2:tpoint):tpoint;
function norm_vect(p:tpoint):tpoint;
function trans(v1,v2:tpoint):ttrans_rs;
function perp_point_line(p:tpoint;l:tline):tpoint;
function dist_point_line(p:tpoint;l:tline):float3d;
function point_in_box(p:tpoint;x1,y1,z1,x2,y2,z2:float3d):io_rel;
function tri_in_box(t:ptriangle;x1,y1,z1,x2,y2,z2:float3d):io_rel;
function norm_2_pane(p,n:tpoint):tpane;
function tri_2_pane(t:ptriangle):tpane;
function pane_diff(e:tpane;p:tpoint):float3d;
function pane_X_line(e:tpane;p1,p2:tpoint;outside:boolean):tX_point;
function line_X_line(p11,p12,p21,p22:tpoint):tX_point;
function pane_norm(e:tpane):tpoint;
function line_tri_line(l:pline;t:ptriangle):ttri_tri_line;
function tri_tri_line(t1,t2:ptriangle):ttri_tri_line;
function point_in_tri(p:tpoint;t:ptriangle):boolean;
function point_in_poly(p:tpoint;no_points:poly_idx3d;poly:ppoly):boolean;
function tri_area(t:ptriangle):float3d;
function points_2_line(p1,p2:tpoint):tline;
function point_on_line(p:tpoint;l:tline;outside:boolean):boolean;
function point_on_line(p:tpoint;l:tline):boolean;
function tri_norm(t:ptriangle):tline;
function tri_tri_in_pane(t1,t2:ptriangle):ttri_tri_rel;
function triangle_triangle(t1,t2:ptriangle):ttrianglerelation;
function split_tri(t:ptriangle;p:tpoint;lineonly:boolean):tsplit_tri;
function split_tri(t:ptriangle;p:tpoint):tsplit_tri;
function tri_has_side(t:ptriangle;l:pline):boolean;
function split_line(t:ptriangle;l:pline;p:tpoint):tsplit_tri;
procedure inside_out(t:ptriangle);

function create_poly(no_points:poly_idx3d;copy_poly:ppoly):ppoly;
function create_poly(no_points:poly_idx3d):ppoly;
function free_poly(poly:ppoly;no_points:poly_idx3d):ppoly;
procedure insert_poly_point(var no_points:poly_idx3d; var poly:ppoly; p:tpoint);

const eps:float3d=0.00001;
			TriLog:boolean=false;

implementation

constructor tshape_3d.init(size:tri_idx3d;nam:string);
begin
	inherited init;
	if size>sizeof(ttrianglearray) div sizeof(ptriangle) then
	  size:=sizeof(ttrianglearray) div sizeof(ptriangle)
	else
		idx_max:=size;
	count:=0;
	name:=nam;
	sort:=none_;
	getmem(tr,idx_max*sizeof(ptriangle));
end;

constructor tshape_3d.init(size:tri_idx3d);
begin
	init(size,'');
end;

function tshape_3d.ins_triangle(idx:tri_idx3d;t:ptriangle):boolean;
var i:tri_idx3d;
begin
	if (count<idx_max) and (idx<=count) then begin
		if idx+1<count then
			for i:=count-1 downto idx do
				tr^[i+1]:=tr^[i];
		tr^[idx]:=t;
		inc(count);
		ins_triangle:=true;
	end else
		ins_triangle:=false;
end;

function tshape_3d.ins_triangle(t:ptriangle):boolean;
begin
	ins_triangle:=ins_triangle(count,t);
end;

function tshape_3d.del_triangle(idx:tri_idx3d):boolean;
var i:tri_idx3d;
begin
	if (count>0) and (idx<count) then begin
		dispose(tr^[idx]);
		tr^[idx]:=NIL;
		if idx+1<count then
			for i:=idx to count-2 do
				tr^[i]:=tr^[i+1];
		dec(count);
	  del_triangle:=true;
	end else
	  del_triangle:=false;
end;

procedure tshape_3d.del_triangles(x1,y1,z1,x2,y2,z2:float3d;rel:sio_rel);
var i:tri_idx3d;
		remove:boolean;
		r:io_rel;
begin
	if (count>0) and (rel<>[]) then begin
{$ifdef LOG}
		if TriLog then begin		
		  writeln('Dreiecke loeschen von Liste ',name,':');
 		  writeln('Pruefquader: (',x1:10:3,y1:10:3,z1:10:3,') (',x2:10:3,y2:10:3,z2:10:3,')');
 		end;
{$endif}
	  for i:=count-1 downto 0 do
	    if assigned(tr^[i]) then begin
    	  r:=tri_in_box(tr^[i],x1,y1,z1,x2,y2,z2);
    	  remove:=r in rel;
	    	if remove then begin
{$ifdef LOG}		
      		if TriLog then begin		
    	    	writeln(i:6,': ',tr^[i]^.p1.x:8:3,tr^[i]^.p1.y:8:3,tr^[i]^.p1.z:8:3);
    	    	writeln(tr^[i]^.p2.x:16:3,tr^[i]^.p2.y:8:3,tr^[i]^.p2.z:8:3);
    	    	writeln(tr^[i]^.p3.x:16:3,tr^[i]^.p3.y:8:3,tr^[i]^.p3.z:8:3);
  	    		case r of
  	    			io_in:writeln('innerhalb');
  	    			io_out:writeln('ausserhalb');
  	    			io_border:writeln('in Flaeche');
  	    			io_split:writeln('geteilt');
  	    		end;
  	    	end;
{$endif}
					del_triangle(i);
	    	end;
	    end;
	end;
end;

function tshape_3d.get_triangle(idx:tri_idx3d):ptriangle;
begin
  if idx<count then
    get_triangle:=tr^[idx]
  else
		get_triangle:=NIL;
end;

procedure tshape_3d.list_triangles(x1,y1,z1,x2,y2,z2:float3d;rel:sio_rel);
var i:tri_idx3d;
		list:boolean;
		r:io_rel;
begin
	if count>0 then begin
		writeln('Liste ',name,':');
  	if rel<>[] then
  		writeln('Pruefquader: (',x1:10:3,y1:10:3,z1:10:3,') (',x2:10:3,y2:10:3,z2:10:3,')');
	  for i:=0 to count-1 do
	    if assigned(tr^[i]) then begin
	    	if rel=[] then
	    		list:=true
	    	else begin
	    	  r:=tri_in_box(tr^[i],x1,y1,z1,x2,y2,z2);
	    	  list:=r in rel;
	    	end;
	    	if list then begin
  	    	writeln(i:6,': ',tr^[i]^.p1.x:8:3,tr^[i]^.p1.y:8:3,tr^[i]^.p1.z:8:3);
  	    	writeln(tr^[i]^.p2.x:16:3,tr^[i]^.p2.y:8:3,tr^[i]^.p2.z:8:3);
  	    	writeln(tr^[i]^.p3.x:16:3,tr^[i]^.p3.y:8:3,tr^[i]^.p3.z:8:3);
  	    	if rel<>[] then
  	    		case r of
  	    			io_in:writeln('innerhalb');
  	    			io_out:writeln('ausserhalb');
  	    			io_border:writeln('in Flaeche');
  	    			io_split:writeln('geteilt');
  	    		end;
	    	end;
	    end;
	end else
		writeln('Liste ',name,' leer:');
end;

procedure tshape_3d.list_triangles(x,y,z:float3d;rel:sio_rel);
var norm:tline;
		n_p_ref:float3d;
    i:tri_idx3d;
		list:boolean;
		r:io_rel;
begin
	if count>0 then begin
		writeln('Liste ',name,':');
  	if rel<>[] then
  		writeln('Pruefpunkt: (',x:10:3,y:10:3,z:10:3,')');
	  for i:=0 to count-1 do
	    if assigned(tr^[i]) then begin
	    	if rel=[] then
	    		list:=true
	    	else begin
	    	  norm:=tri_norm(tr^[i]);
	    	  n_p_ref:=vec_p_vec(point_point(norm.p1,norm.p2),point_point(norm.p1,point(x,y,z)));
	    	  if n_p_ref=0 then
	    	    r:=io_border
	    	  else
	    	    if n_p_ref>0 then
	    	      r:=io_out
	    	    else
	    	      r:=io_in;
	    	  list:=r in rel;
	    	end;
	    	if list then begin
  	    	writeln(i:6,': ',tr^[i]^.p1.x:8:3,tr^[i]^.p1.y:8:3,tr^[i]^.p1.z:8:3);
  	    	writeln(tr^[i]^.p2.x:16:3,tr^[i]^.p2.y:8:3,tr^[i]^.p2.z:8:3);
  	    	writeln(tr^[i]^.p3.x:16:3,tr^[i]^.p3.y:8:3,tr^[i]^.p3.z:8:3);
  	    	if rel<>[] then
  	    		case r of
  	    			io_in:writeln('Referenzpunkt unter(inner)halb');
  	    			io_out:writeln('Referenzpunkt ober(ausser)halb');
  	    			io_border:writeln('Referenzpunkt in Flaeche');
  	    		end;
	    	end;
	    end;
	end else
		writeln('Liste ',name,' leer:');
end;

procedure tshape_3d.list_triangles;
begin
  list_triangles(0,0,0,0,0,0,[]);
end;

procedure tshape_3d.inside2out(mask:byte);
var i:tri_idx3d;
begin
	if count>0 then
	  for i:=0 to count-1 do
	    if assigned(tr^[i]) then
	      if (mask and tr^[i]^.flag)=mask then begin
	    		inside_out(tr^[i]);
	    		tr^[i]^.flag:=tr^[i]^.flag-mask;
	    	end;
end;

procedure tshape_3d.inside2out;
begin
  inside2out(0);
end;

procedure tshape_3d.change_size(sm:size_mode);
var tr_:ptrianglearray;
		idx_max_:tri_idx3d;
begin
	case sm of
		cl_minimize:begin
									getmem(tr_,count*sizeof(ptriangle));
									idx_max_:=count;
{$ifdef LOG}
  								if TriLog then
   									writeln('Feldgroesse wird minimiert (',count,' Eintraege)');
{$endif}
								end;
		cl_maximize:begin
									getmem(tr_,sizeof(ttrianglearray));
									idx_max_:=sizeof(ttrianglearray) div sizeof(ptriangle);
{$ifdef LOG}
  								if TriLog then
   									writeln('Feldgroesse wird maximiert (',sizeof(ttrianglearray) div sizeof(ptriangle),' Eintraege)');
{$endif}
								end;
	end;
	move(tr^,tr_^,count*sizeof(ptriangle));
	freemem(tr,idx_max*sizeof(ptriangle));
	tr:=tr_;
	idx_max:=idx_max_;
end;

function tshape_3d.clone_shape(nam:string;sm:size_mode):pshape_3d;
var i:tri_idx3d;
    t:ptriangle;
{$ifndef FKT_RESULT}
    result:pshape_3d;
{$endif}
begin
	case sm of
		cl_minimize:new(result,init(count,nam));
	  cl_exact:new(result,init(idx_max,nam));
	  cl_maximize:new(result,init(sizeof(ttrianglearray) div sizeof(ptriangle),nam));
	end;
	if count>0 then
	  for i:=0 to count-1 do
	    if assigned(tr^[i]) then begin
	    	new(t);
	    	dup_triangle(i,t^);
	    	result^.ins_triangle(t);
	    end;
{$ifndef FKT_RESULT}
	clone_shape:=result;
{$endif}
end;

function tshape_3d.clone_shape(nam:string):pshape_3d;
{$ifndef FKT_RESULT}
var result:pshape_3d;
{$endif}
begin
	result:=clone_shape(nam,cl_exact);
{$ifndef FKT_RESULT}
	clone_shape:=result;
{$endif}
end;

procedure tshape_3d.dup_triangle(i:tri_idx3d;var t:ttriangle);
begin
 	t:=tr^[i]^;
end;

function tshape_3d.include_triangles(s:pshape_3d):boolean;
var i:tri_idx3d;
    t:ptriangle;
{$ifndef FKT_RESULT}
    result:boolean;
{$endif}
begin
	if count+s^.count<=idx_max then begin
	  if s^.count>0 then
			for i:=0 to s^.count-1 do
			  if assigned(s^.tr^[i]) then begin
			  	new(t);
		    	s^.dup_triangle(i,t^);
			  	ins_triangle(t);
			  end;
	  result:=true;
	end else
		result:=false;
{$ifndef FKT_RESULT}
	include_triangles:=result;
{$endif}
end;

function tshape_3d.break_triangles(t:ptriangle):boolean;
var i,j:tri_idx3d;
		t_s:tsplit_tri;
		l:tline;
{$ifndef FKT_RESULT}
		result:boolean;
{$endif}
begin
	result:=true;
	i:=0;
	while (i<count) and result do begin
		l:=points_2_line(t^.p1,t^.p2);
		t_s:=split_line(get_triangle(i),@l,t^.p3);
		if t_s.ok then begin
			for j:=1 to 3 do
			  if assigned(t_s.t[j]) then
			    if not ins_triangle(t_s.t[j]) then
			      result:=false;
			if result then
				del_triangle(i);
		end else
		  inc(i);
	end;
{$ifndef FKT_RESULT}
	break_triangles:=result;
{$endif}
end;

function tshape_3d.sum_area:float3d;
var i:tri_idx3d;
{$ifndef FKT_RESULT}
		result:float3d;
{$endif}
begin
	result:=0;
	if count>0 then
  	for i:=0 to count-1 do begin
  		result:=result+tri_area(tr^[i]);
  	end;
{$ifndef FKT_RESULT}
	sum_area:=result;
{$endif}
end;

destructor tshape_3d.done;
var i:tri_idx3d;
begin
	if idx_max>0 then begin
		for i:=0 to count-1 do
			if assigned(tr^[i]) then
  			dispose(tr^[i]);
  	freemem(tr,idx_max*sizeof(ptriangle));
  end;
	inherited done;
end;

function min(v1,v2:float3d):float3d;
{Minimalwert zweier Variablenwerte}
begin
	if v1<v2 then
	  min:=v1
	else
	  min:=v2;
end;

function max(v1,v2:float3d):float3d;
{Maximalwert zweier Variablenwerte}
begin
	if v1>v2 then
	  max:=v1
	else
	  max:=v2;
end;

procedure swap(var a,b:float3d);
{Austausch zweier Variablenwerte}
var c:float3d;
begin
  c:=b;
  b:=a;
  a:=c;
end;

procedure swap(var a,b:tri_idx3d);
{Austausch zweier Variablenwerte}
var c:tri_idx3d;
begin
  c:=b;
  b:=a;
  a:=c;
end;

function point(x,y,z:float3d):tpoint;
{erzeugt einen Punkt aus den 3 Koordinatenwerten}
{$ifndef FKT_RESULT}
var result:tpoint;
{$endif}
begin
	result.x:=x;
	result.y:=y;
	result.z:=z;
{$ifndef FKT_RESULT}
	point:=result;
{$endif}
end;

function same_point(p1,p2:tpoint):boolean;
{prueft auf Gleichheit der Punkte}
begin
  same_point:=(p1.x+eps>=p2.x) and (p1.x-eps<=p2.x) and
  						(p1.y+eps>=p2.y) and (p1.y-eps<=p2.y) and
  						(p1.z+eps>=p2.z) and (p1.z-eps<=p2.z);
end;

function point_point(p1,p2:tpoint):tpoint;
{Vektor zwischen zwei Punkten}
{$ifndef FKT_RESULT}
var result:tpoint;
{$endif}
begin
	result.x:=p2.x-p1.x;
	result.y:=p2.y-p1.y;
	result.z:=p2.z-p1.z;
{$ifndef FKT_RESULT}
	point_point:=result;
{$endif}
end;

function vec_len(v:tpoint):float3d;
{Vektorbetrag}
begin
	vec_len:=sqrt(v.x*v.x+v.y*v.y+v.z*v.z);
end;

function vec_add(p1,p2:tpoint):tpoint;
{Vektoraddition}
{$ifndef FKT_RESULT}
var result:tpoint;
{$endif}
begin
	result.x:=p1.x+p2.x;
	result.y:=p1.y+p2.y;
	result.z:=p1.z+p2.z;
{$ifndef FKT_RESULT}
	vec_add:=result;
{$endif}
end;

function vec_scale(v:tpoint;s:float3d):tpoint;
{Skalierung eines Vektors}
{$ifndef FKT_RESULT}
var result:tpoint;
{$endif}
begin
	result.x:=v.x*s;
	result.y:=v.y*s;
	result.z:=v.z*s;
{$ifndef FKT_RESULT}
	vec_scale:=result;
{$endif}
end;

function vec_p_vec(v1,v2:tpoint):float3d;
{Punktprodukt zweier Vektoren}
{$ifndef FKT_RESULT}
var result:float3d;
{$endif}
begin
	result:=v1.x*v2.x+v1.y*v2.y+v1.z*v2.z;
{$ifndef FKT_RESULT}
	vec_p_vec:=result;
{$endif}
end;

function vec_X_vec(v1,v2:tpoint):tpoint;
{Vektorprodukt}
{$ifndef FKT_RESULT}
var result:tpoint;
{$endif}
begin
	result.x:=v1.y*v2.z-v1.z*v2.y;
	result.y:=v1.z*v2.x-v1.x*v2.z;
	result.z:=v1.x*v2.y-v1.y*v2.x;
{$ifndef FKT_RESULT}
	vec_X_vec:=result;
{$endif}
end;

function norm_vect(p:tpoint):tpoint;
{normiert einen Vektor}
var l:float3d;
{$ifndef FKT_RESULT}
		result:tpoint;
{$endif}
begin
	l:=vec_len(p);
	if l<>0 then
		result:=vec_scale(p,1/l)
	else
		result:=vec_scale(p,0);
{$ifndef FKT_RESULT}
	norm_vect:=result;
{$endif}
end;

function trans(v1,v2:tpoint):ttrans_rs;
{Berechnung der Transformationen Rotation und Skalierung, die von v1 zu v2 fhren (gleicher Ursprung vorausgesetzt)}
{$ifndef FKT_RESULT}
var result:ttrans_rs;
{$endif}
begin
{$ifdef LOG}		
	if TriLog then
		writeln('Transformation von <',v1.x:0:3,',',v1.y:0:3,',',v1.z:0:3,'> zu <',v2.x:0:3,',',v2.y:0:3,',',v2.z:0:3,'>');
{$endif}
	result.axis:=vec_X_vec(v1,v2);
	result.angle:=0;
	if (vec_p_vec(v1,v2)<>0) and (vec_len(result.axis)>0) then begin
{$ifdef LOG}
		if TriLog then
	 	  writeln('Punktprodukt: ',vec_p_vec(v1,v2):0:3,', Vektorprodukt: |<',result.axis.x:0:3,
 		  	',',result.axis.y:0:3,',',result.axis.z:0:3,'>| =',vec_len(result.axis):0:3);
{$endif}
	  result.angle:=arctan(vec_len(result.axis)/vec_p_vec(v1,v2));
	  if result.angle<0 then
	  	result.angle:=result.angle+pi;
	end else
	  if vec_len(result.axis)>0 then
		  result.angle:=pi/2;
{$ifdef LOG}
	if TriLog then
	  writeln('Winkel: ',result.angle:0:3,' (',result.angle*180/pi:0:3,')');
{$endif}
	result.scale:=1;
	if (vec_len(v1)>0) then
	  result.scale:=vec_len(v2)/vec_len(v1);
{$ifndef FKT_RESULT}
	trans:=result;
{$endif}
end;

function perp_point_line(p:tpoint;l:tline):tpoint;
{Lotpunkt von einem Punkt auf eine Linie}
{$ifndef FKT_RESULT}
var result:tpoint;
{$endif}
begin
	result:=norm_vect(point_point(l.p1,l.p2));
	result:=vec_add(l.p1,vec_scale(result,vec_p_vec(result,point_point(l.p1,p))));
{$ifndef FKT_RESULT}
	perp_point_line:=result;
{$endif}
end;

function dist_point_line(p:tpoint;l:tline):float3d;
{Abstand zwischen Punkt und Linie}
{$ifndef FKT_RESULT}
var result:float3d;
{$endif}
begin
	result:=vec_len(point_point(p,perp_point_line(p,l)));
{$ifndef FKT_RESULT}
	dist_point_line:=result;
{$endif}
end;

function point_in_box(p:tpoint;x1,y1,z1,x2,y2,z2:float3d):io_rel;
{ermittelt, ob ein Punkt innerhalb, auf einer Seitenflaeche oder
ausserhalb eines vorgegebenen Quaders liegen}
var p1,p2:tpoint;
{$ifndef FKT_RESULT}
		result:io_rel;
{$endif}
begin
	result:=io_border;
	if x2<x1 then begin
		p1.x:=x2; p2.x:=x1;
	end else begin
		p1.x:=x1; p2.x:=x2;
	end;
	if y2<y1 then begin
		p1.y:=y2; p2.y:=y1;
	end else begin
		p1.y:=y1; p2.y:=y2;
	end;
	if z2<z1 then begin
		p1.z:=z2; p2.z:=z1;
	end else begin
		p1.z:=z1; p2.z:=z2;
	end;
	if (p.x<p1.x-eps) or (p.x>p2.x+eps) or (p.y<p1.y-eps) or (p.y>p2.y+eps) or (p.z<p1.z-eps) or (p.Z>p2.z+eps) then
		result:=io_out
	else
		if ((p.x>p1.x+eps) and (p.x<p2.x-eps)) and ((p.y>p1.y+eps) and (p.y<p2.y-eps)) and ((p.z>p1.z+eps) and (p.z<p2.z-eps)) then
			result:=io_in;
{$ifndef FKT_RESULT}
	point_in_box:=result;
{$endif}
end;

function tri_in_box(t:ptriangle;x1,y1,z1,x2,y2,z2:float3d):io_rel;
{ermittelt, ob die Eck- und Seitenmittelpunkte punkte eines Dreiecks innerhalb, auf den Seitenflaechen,
ausserhalb oder teilweise innerhalb und ausserhalb eines vorgegebenen Quaders liegen}
var	p:tpoint;
		i:byte;
{$ifndef FKT_RESULT}
		result:io_rel;
{$endif}
begin
  result:=io_border;
  for i:=1 to 6 do begin
  	case i of
  		1:p:=t^.p1;
  		2:p:=vec_scale(vec_add(t^.p1,t^.p2),0.5);
  		3:p:=t^.p2;
  		4:p:=vec_scale(vec_add(t^.p2,t^.p3),0.5);
  		5:p:=t^.p3;
  		6:p:=vec_scale(vec_add(t^.p3,t^.p1),0.5);
  	end;
  	if result in [io_border,io_in,io_out] then begin
  		case point_in_box(p,x1,y1,z1,x2,y2,z2) of
  			io_in:if result=io_out then result:=io_split else result:=io_in;
  			io_out:if result=io_in then result:=io_split else result:=io_out;
  		end;
  	end;
  end;
{$ifndef FKT_RESULT}
  tri_in_box:=result;
{$endif}
end;

function norm_2_pane(p,n:tpoint):tpane;
{berechnet Ebene aus Basispunkt und Normalenrichtung}
{$ifndef FKT_RESULT}
var result:tpane;
{$endif}
begin
	with n do begin
	  result.x:=x;
		result.y:=y;
		result.z:=z;
	end;
	result.c:=-(p.x*result.x+p.y*result.y+p.z*result.z);
{$ifndef FKT_RESULT}
	norm_2_pane:=result;
{$endif}
end;

function tri_2_pane(t:ptriangle):tpane;
{Normalenform der Ebenengleichung zum Dreieck}
{$ifndef FKT_RESULT}
var result:tpane;
{$endif}
begin
  result:=norm_2_pane(t^.p1,vec_X_vec(point_point(t^.p1,t^.p2),point_point(t^.p1,t^.p3)));
{$ifndef FKT_RESULT}
	tri_2_pane:=result;
{$endif}
end;

function pane_diff(e:tpane;p:tpoint):float3d;
{Abstand zwischen Ebene und Punkt +/-}
begin
  pane_diff:=p.x*e.x+p.y*e.y+p.z*e.z+e.c;
end;

function pane_X_line(e:tpane;p1,p2:tpoint;outside:boolean):tX_point;
{Schnittpunkt von Linie und Ebene}
var a1,a2:float3d;
{$ifndef FKT_RESULT}
		result:tX_point;
{$endif}
begin
  result.ok:=false;
  result.in_pane:=false;
	a1:=pane_diff(e,p1);
	a2:=pane_diff(e,p2);
{$ifdef LOG}
	if TriLog then begin		
  	writeln('Bezugsebene:',e.x:10:4,e.y:10:4,e.z:10:4,e.c:10:4);
  	writeln('Abstnde: Normalenfupunkt:',a1:10:4,' Spitze:',a2:10:4);
	end;
{$endif}
	if a1<>a2 then begin
 	  result.ok:=true;
 	  with point_point(p1,p2) do begin
 			result.p.x:=p1.x+a1/(a1-a2)*x;
 			result.p.y:=p1.y+a1/(a1-a2)*y;
 			result.p.z:=p1.z+a1/(a1-a2)*z;
 		end;
{$ifdef LOG}
		if TriLog then
	  	writeln('Durchstopunkt:',result.p.x:10:4,result.p.y:10:4,result.p.z:10:4);
{$endif}
 		if (not outside) and ((a1/(a1-a2)<0) or (a1/(a1-a2)>1)) then
 			if not (same_point(p1,result.p) or same_point(p2,result.p)) then
 				result.ok:=false;
 	end;
{$ifndef FKT_RESULT}
	pane_X_line:=result;
{$endif}
end;

function line_X_line(p11,p12,p21,p22:tpoint):tX_point;
{Schnittpunkt zweier Strecken im Raum}
var	n:tpoint;
		e:tpane;
{$ifndef FKT_RESULT}
		result:tX_point;
{$endif}
begin
	n:=vec_X_vec(point_point(p11,p12),vec_X_vec(point_point(p11,p12),point_point(p21,p22)));
	e:=norm_2_pane(p11,n);
	result:=pane_X_line(e,p21,p22,false);
	if result.ok then
	  result.ok:=point_on_line(result.p,points_2_line(p11,p12),false);
{$ifndef FKT_RESULT}
	line_X_line:=result;
{$endif}
end;

function pane_norm(e:tpane):tpoint;
{Normalenvektor einer Ebene}
{$ifndef FKT_RESULT}
var result:tpoint;
{$endif}
begin
  result.x:=e.x;
  result.y:=e.y;
  result.z:=e.z;
{$ifndef FKT_RESULT}
  pane_norm:=result;
{$endif}
end;

function line_tri_line(l:pline;t:ptriangle):ttri_tri_line;
{Im Dreieck liegender Anteil einer Linie}
var p:array[1..5] of tX_point;
		idx,jdx:byte;
		pidx:array[1..2] of byte;
{$ifndef FKT_RESULT}
		result:ttri_tri_line;
{$endif}		
begin
	result.ok:=false;
 	p[4].p:=l^.p1;p[4].ok:=true;
 	p[5].p:=l^.p2;p[5].ok:=true;
 	p[1]:=line_X_line(l^.p1,l^.p2,t^.p1,t^.p2);
 	p[2]:=line_X_line(l^.p1,l^.p2,t^.p2,t^.p3);
 	p[3]:=line_X_line(l^.p1,l^.p2,t^.p3,t^.p1);
	for idx:=1 to 5 do
		if p[idx].ok then
  		if not point_in_tri(p[idx].p,t) then
  		  p[idx].ok:=false;
	for idx:=1 to 4 do
  	if p[idx].ok then
    	for jdx:=idx+1 to 5 do
      	if p[jdx].ok then
      		if same_point(p[idx].p,p[jdx].p) then
      		  p[jdx].ok:=false;
  idx:=0;jdx:=1;
  pidx[1]:=0;pidx[2]:=0;
  while ((pidx[1]=0) or (pidx[2]=0)) and (idx<5) do begin
    inc(idx);
  	if p[idx].ok then begin
  	  if jdx<=2 then
  	  	pidx[jdx]:=idx;
  	  inc(jdx);
  	end;
  end;
  if jdx=3 then begin
    result.p1:=p[pidx[1]].p;
    result.p2:=p[pidx[2]].p;
    result.ok:=true;
  end;
{$ifndef FKT_RESULT}
  line_tri_line:=result;
{$endif}
end;

function tri_tri_line(t1,t2:ptriangle):ttri_tri_line;
{Schnittlinie zweier Dreiecke}
var p:array[1..2,1..3] of tX_point;
		q:array[1..2,1..2] of pX_point;
    sp1,sp2,inp1,inp2:byte;
		i,j,k:byte;
		dist:array[1..3] of float3d;
		idx:array[1..2] of byte;
{$ifndef FKT_RESULT}
		result:ttri_tri_line;
{$endif}
begin
	result.ok:=false;
	sp1:=0;inp1:=0;
	sp2:=0;inp2:=0;
  p[1,1]:=pane_X_Line(tri_2_pane(t2),t1^.p1,t1^.p2,false);
  p[1,2]:=pane_X_Line(tri_2_pane(t2),t1^.p2,t1^.p3,false);
  p[1,3]:=pane_X_Line(tri_2_pane(t2),t1^.p3,t1^.p1,false);
{$ifdef Log}
	if TriLog then begin		
  	if p[1,1].ok then
  		Write('<1,1:',p[1,1].p.x:0:4,',',p[1,1].p.y:0:4,',',p[1,1].p.z:0:4,'>');
  	if p[1,2].ok then
  		Write('<1,2:',p[1,2].p.x:0:4,',',p[1,2].p.y:0:4,',',p[1,2].p.z:0:4,'>');
  	if p[1,3].ok then
  		Write('<1,3:',p[1,3].p.x:0:4,',',p[1,3].p.y:0:4,',',p[1,3].p.z:0:4,'>');
  	WriteLN();
	end;
{$endif}
  if p[1,1].ok and p[1,2].ok and p[1,3].ok then
    if same_point(t1^.p2,p[1,1].p) and same_point(t1^.p2,p[1,2].p) then begin
    	p[1,2].ok:=false; p[1,2].in_pane:=false;
    end else
      if same_point(t1^.p1,p[1,1].p) and same_point(t1^.p1,p[1,3].p) then begin
      	p[1,3].ok:=false; p[1,3].in_pane:=false;
      end else
        if same_point(t1^.p3,p[1,2].p) and same_point(t1^.p3,p[1,3].p) then begin
        	p[1,3].ok:=false; p[1,3].in_pane:=false;
        end;
  for i:=1 to 3 do begin
	  if p[1,i].ok then inc(sp1);
	  if p[1,i].in_pane then inc(inp1);
	end;
{$ifdef Log}
	if TriLog then		
		WriteLN('Punkte ok/in Flaeche:',sp1,'/',inp1);
{$endif}
	p[2,1]:=pane_X_Line(tri_2_pane(t1),t2^.p1,t2^.p2,false);
	p[2,2]:=pane_X_Line(tri_2_pane(t1),t2^.p2,t2^.p3,false);
	p[2,3]:=pane_X_Line(tri_2_pane(t1),t2^.p3,t2^.p1,false);
{$ifdef Log}
	if TriLog then begin		
  	if p[2,1].ok then
  		Write('<2,1:',p[2,1].p.x:0:4,',',p[2,1].p.y:0:4,',',p[2,1].p.z:0:4,'>');
  	if p[2,2].ok then
  		Write('<2,2:',p[2,2].p.x:0:4,',',p[2,2].p.y:0:4,',',p[2,2].p.z:0:4,'>');
  	if p[2,3].ok then
  		Write('<2,3:',p[2,3].p.x:0:4,',',p[2,3].p.y:0:4,',',p[2,3].p.z:0:4,'>');
  	WriteLN();
	end;
{$endif}
  if p[2,1].ok and p[2,2].ok and p[2,3].ok then
    if same_point(t2^.p2,p[2,1].p) and same_point(t2^.p2,p[2,2].p) then begin
    	p[2,2].ok:=false; p[2,2].in_pane:=false;
    end else
      if same_point(t2^.p1,p[2,1].p) and same_point(t2^.p1,p[2,3].p) then begin
      	p[2,3].ok:=false; p[2,3].in_pane:=false;
      end else
        if same_point(t2^.p3,p[2,2].p) and same_point(t2^.p3,p[2,3].p) then begin
        	p[2,3].ok:=false; p[2,3].in_pane:=false;
        end;
  for i:=1 to 3 do begin
	  if p[2,i].ok then inc(sp2);
	  if p[2,i].in_pane then inc(inp2);
	end;
{$ifdef Log}
	if TriLog then
		WriteLN('Punkte ok/in Flaeche:',sp2,'/',inp2);
{$endif}
	if (sp1=2) and (sp2=2) and ((sp1-inp1>0) or (sp2-inp2>0)) then begin
		for i:=1 to 2 do begin
		  k:=1;
		  for j:=1 to 3 do
		    if p[i,j].ok then begin
		      q[i,k]:=@p[i,j];
		      inc(k);
		    end;
		end;
		result.p1:=q[1,1]^.p;
		result.p2:=q[1,2]^.p;
 		dist[1]:=vec_p_vec(point_point(q[1,1]^.p,q[1,2]^.p),point_point(q[1,1]^.p,q[1,2]^.p));
 		dist[2]:=vec_p_vec(point_point(q[1,1]^.p,q[2,1]^.p),point_point(q[1,1]^.p,q[1,2]^.p));
 		dist[3]:=vec_p_vec(point_point(q[1,1]^.p,q[2,2]^.p),point_point(q[1,1]^.p,q[1,2]^.p));
{$ifdef Log}
		if TriLog then
			WriteLN('Abstnde:',dist[1]:0:4,',',dist[2]:0:4,',',dist[3]:0:4);
{$endif}
 		idx[1]:=1;
 		idx[2]:=2;
 		if dist[3]<dist[2] then begin
 		  swap(dist[2],dist[3]);
    		idx[1]:=2;
 		  idx[2]:=1;
 		end;
 		if (dist[3]>0) and (dist[2]<dist[1]) then begin
 		  if dist[2]>0 then
 		    result.p1:=q[2,idx[1]]^.p;
 		  if dist[3]<dist[1] then
 		    result.p2:=q[2,idx[2]]^.p;
 		  result.ok:=true;
 		end;
	end;
{
Das Vektorprodukt (Schnittlinie L X Flaechennormale des Dreiecks t1) zeigt in Richtung des
ueber t2 liegenden (aussenliegenden) Flaechenanteiles von t1
Das Vektorprodukt (Schnittlinie L X Flaechennormale des Dreiecks t2) zeigt in Richtung des
unter t1 liegenden (innenliegenden) Flaechenanteiles von t2

(N2 (Flaechennormale des Dreiecks t1) x N1) . L > 0
}
	if result.ok then
  	if vec_p_vec(point_point(result.p1,result.p2),
  		vec_X_vec(
  		vec_X_vec(point_point(t2^.p1,t2^.p2),point_point(t2^.p1,t2^.p3)),
  		vec_X_vec(point_point(t1^.p1,t1^.p2),point_point(t1^.p1,t1^.p3))
  		)) < 0
  	then begin
  		swap(result.p1.x,result.p2.x);
  		swap(result.p1.y,result.p2.y);
  		swap(result.p1.z,result.p2.z);
  	end;
{$ifndef FKT_RESULT}
	tri_tri_line:=result;
{$endif}
end;

function point_in_tri(p:tpoint;t:ptriangle):boolean;
{prueft, ob Punkt in Dreieck liegt}
var e:tpane;
		n:tpoint;
{$ifndef FKT_RESULT}
		result:boolean;
{$endif}
begin
	e:=tri_2_pane(t);
	n:=pane_norm(e);
	result:=(vec_p_vec(point_point(t^.p1,p),vec_X_vec(n,point_point(t^.p1,t^.p2)))>=0) and
			(vec_p_vec(point_point(t^.p2,p),vec_X_vec(n,point_point(t^.p2,t^.p3)))>=0) and
			(vec_p_vec(point_point(t^.p3,p),vec_X_vec(n,point_point(t^.p3,t^.p1)))>=0);
	if not result then
	  if (point_on_line(p,points_2_line(t^.p1,t^.p2))) or
	  	(point_on_line(p,points_2_line(t^.p2,t^.p3))) or
	  	(point_on_line(p,points_2_line(t^.p3,t^.p1)))
	  then
	  	result:=true;
{$ifndef FKT_RESULT}
	point_in_tri:=result;
{$endif}
end;

function point_in_poly(p:tpoint;no_points:poly_idx3d;poly:ppoly):boolean;
{ermittelt, ob Punkt im Polygon liegt}
var i:poly_idx3d;
		x:poly_idx3d;
		pa:tpoint;
		d,d_akt:float3d;
{$ifndef FKT_RESULT}
		result:boolean;
{$endif}
begin
	result:=false;
	if same_point(poly^[0],poly^[no_points-1]) then begin
		x:=0;
		d:=0;
		for i:=1 to no_points-1 do begin
			d_akt:=vec_len(point_point(poly^[0],poly^[i]));
			if d_akt>d then begin
				pa:=vec_add(poly^[0],vec_scale(point_point(poly^[0],poly^[i]),2));
				d:=d_akt;
			end;
		end;
		if d>0 then begin
  		for i:=0 to no_points-2 do
    		if line_X_line(p,pa,poly^[i],poly^[i+1]).ok then
    			inc(x);
    	result:=(x mod 2)=1;
  	end;
	end;
{$ifndef FKT_RESULT}
	point_in_poly:=result;
{$endif}
end;

function tri_area(t:ptriangle):float3d;
{Flaeche eines Dreiecks}
{$ifndef FKT_RESULT}
var result:float3d;
{$endif}
begin
	result:=vec_len(vec_X_vec(point_point(t^.p1,t^.p2),point_point(t^.p1,t^.p3)))*0.5;
{$ifndef FKT_RESULT}
	tri_area:=result;
{$endif}
end;

function points_2_line(p1,p2:tpoint):tline;
{erzeugt aus 2 Punkten eine Linie}
{$ifndef FKT_RESULT}
var result:tline;
{$endif}
begin
	result.p1:=p1;
	result.p2:=p2;
{$ifndef FKT_RESULT}
	points_2_line:=result;
{$endif}
end;

function point_on_line(p:tpoint;l:tline;outside:boolean):boolean;
{prueft, ob Punkt auf Linie liegt}
{$ifndef FKT_RESULT}
var result:boolean;
{$endif}
begin
	result:=false;
	if outside or ((p.x>=min(l.p1.x,l.p2.x)-eps) and (p.x<=max(l.p1.x,l.p2.x)+eps)
	  and (p.y>=min(l.p1.y,l.p2.y)-eps) and (p.y<=max(l.p1.y,l.p2.y)+eps)
	  and (p.z>=min(l.p1.z,l.p2.z)-eps) and (p.z<=max(l.p1.z,l.p2.z)+eps))
	then
  	result:=dist_point_line(p,l)<=eps;
{$ifndef FKT_RESULT}
  point_on_line:=result;
{$endif}
end;

function point_on_line(p:tpoint;l:tline):boolean;
begin
	point_on_line:=point_on_line(p,l,false);
end;

function tri_norm(t:ptriangle):tline;
{$ifndef FKT_RESULT}
var result:tline;
{$endif}
begin
	result.p1.x:=(t^.p1.x+t^.p2.x+t^.p3.x)/3;
	result.p1.y:=(t^.p1.y+t^.p2.y+t^.p3.y)/3;
	result.p1.z:=(t^.p1.z+t^.p2.z+t^.p3.z)/3;
	with pane_norm(tri_2_pane(t)) do begin
		result.p2.x:=result.p1.x+x;
		result.p2.y:=result.p1.y+y;
		result.p2.z:=result.p1.z+z;
	end;
{$ifdef LOG}
	if TriLog then begin		
  	writeln('Dreiecksnormale:');
  	writeln('Dreieck: (',t^.p1.x:7:3,t^.p1.y:7:3,t^.p1.z:7:3,') (',
  		t^.p2.x:7:3,t^.p2.y:7:3,t^.p2.z:7:3,') (',t^.p3.x:7:3,t^.p3.y:7:3,t^.p3.z:7:3,')');
  	writeln('Normale: (',result.p1.x:7:3,result.p1.y:7:3,result.p1.z:7:3,') (',result.p2.x:7:3,result.p2.y:7:3,result.p2.z:7:3,')');
  	writeln;
	end;
{$endif}
{$ifndef FKT_RESULT}
	tri_norm:=result;
{$endif}
end;

function tri_tri_in_pane(t1,t2:ptriangle):ttri_tri_rel;
{ermittelt, ob Dreiecke exakt in einer Ebene liegen}
var t_1,t_2:ttriangle;
		n1,n2:tpoint;
		p1,p2:tpane;
{$ifndef FKT_RESULT}
		result:ttri_tri_rel;
{$endif}		
begin
  with tri_norm(t1) do
		n1:=norm_vect(point_point(p1,p2));
	with tri_norm(t2) do
		n2:=norm_vect(point_point(p1,p2));
	result:=tt_none;
	if vec_len(vec_X_vec(n1,n2))<eps then begin
		t_1.p1:=vec_add(t1^.p1,vec_scale(n1,eps));
		t_1.p2:=vec_add(t1^.p2,vec_scale(n1,eps));
		t_1.p3:=vec_add(t1^.p3,vec_scale(n1,eps));
		t_2.p1:=vec_add(t1^.p1,vec_scale(n1,-eps));
		t_2.p2:=vec_add(t1^.p2,vec_scale(n1,-eps));
		t_2.p3:=vec_add(t1^.p3,vec_scale(n1,-eps));
		p1:=tri_2_pane(@t_1);
		p2:=tri_2_pane(@t_2);
		if (pane_diff(p1,t2^.p1)<0) and (pane_diff(p1,t2^.p2)<0) and (pane_diff(p1,t2^.p3)<0) and
			(pane_diff(p2,t2^.p1)>0) and (pane_diff(p2,t2^.p2)>0) and (pane_diff(p2,t2^.p3)>0)
		then
			if vec_P_vec(n1,n2)>0 then
			  result:=tt_pane_double
			else
				result:=tt_pane_anti;
	end;
{$ifndef FKT_RESULT}
	tri_tri_in_pane:=result;
{$endif}
end;

function triangle_triangle(t1,t2:ptriangle):ttrianglerelation;
{ermittelt die Beziehung zweier Dreiecksfacetten (welche Seiten sind einander zugewandt?)}
var	l1,l2:tline;
		p:tX_point;
{$ifndef FKT_RESULT}
		result:ttrianglerelation;
{$endif}
begin
  l1:=tri_norm(t1);
{$ifdef LOG}
	if TriLog then
  	writeln('Normale des 1. Dreiecks: (',l1.p1.x:7:3,l1.p1.y:7:3,l1.p1.z:7:3,') (',l1.p2.x:7:3,l1.p2.y:7:3,l1.p2.z:7:3,')');
{$endif}
	p:=pane_X_line(tri_2_pane(t2),l1.p1,l1.p2,true);
{$ifdef LOG}
	if TriLog then		
    if p.ok then
  	  writeln('Durchstich der Normale des 1. Dreiecks durch 2. Dreieck: (',p.p.x:7:3,p.p.y:7:3,p.p.z:7:3,')')
  	else
  	  writeln('kein Durchstich der Normale des 1. Dreiecks durch Dreieck 2');
{$endif}
	result.r:=tt_none;
	result.d:=0;
	if p.ok then
		if point_in_tri(p.p,t2) then begin
{$ifdef LOG}
  		if TriLog then
  			writeln('Durchstich im 2. Dreieck: (',p.p.x:7:3,p.p.y:7:3,p.p.z:7:3,')');
{$endif}
			result.d:=vec_len(point_point(l1.p1,p.p));
			if result.d<eps then
			  result.r:=tri_tri_in_pane(t1,t2);
			if result.r=tt_none then begin
		  	l2:=tri_norm(t2);
{$ifdef LOG}
  			if TriLog then begin		
  				writeln('Normale des 2. Dreiecks: (',l2.p1.x:7:3,l2.p1.y:7:3,l2.p1.z:7:3,') (',l2.p2.x:7:3,l2.p2.y:7:3,l2.p2.z:7:3,')');
  				writeln;
  			end;
{$endif}
  			if vec_p_vec(point_point(l2.p1,l2.p2),point_point(p.p,l1.p1))<0 then
  				if vec_p_vec(point_point(l2.p1,l2.p2),point_point(l1.p1,l1.p2))<0 then begin
  					result.r:=tt_back_2_back;
  					result.d:=-result.d;
  				end else
  				  result.r:=tt_surf_2_back
  			else
  				if vec_p_vec(point_point(l2.p1,l2.p2),point_point(l1.p1,l1.p2))<0 then
  				  result.r:=tt_surf_2_surf
  				else begin
  					result.r:=tt_back_2_surf;
  					result.d:=-result.d;
  				end;
			end else
			  result.d:=0;
		end;
{$ifndef FKT_RESULT}
	triangle_triangle:=result;
{$endif}
end;

function split_tri(t:ptriangle;p:tpoint;lineonly:boolean):tsplit_tri;
{unterteilt eine Dreiecksfacette in Subdreiecke, die einen weiteren Punkt beruecksichtigen}
var i:byte;
		is_same:byte;
{$ifndef FKT_RESULT}
		result:tsplit_tri;
{$endif}
begin
	for i:=1 to 3 do
	  result.t[i]:=NIL;
	result.l:=NIL;
	result.ok:=false;
	is_same:=0;
	if same_point(t^.p1,p) then begin
		if dist_point_line(p,points_2_line(t^.p2,t^.p3))>vec_len(point_point(p,t^.p1)) then
			is_same:=1
		else
			is_same:=5;
	end;
	if same_point(t^.p2,p) then begin
		if dist_point_line(p,points_2_line(t^.p3,t^.p1))>vec_len(point_point(p,t^.p2)) then
			is_same:=2
		else
			is_same:=6;
	end;
	if same_point(t^.p3,p) then begin
		if dist_point_line(p,points_2_line(t^.p1,t^.p2))>vec_len(point_point(p,t^.p3)) then
			is_same:=3
		else
			is_same:=4;
	end;
	if (point_in_tri(p,t)) and (is_same in [0,4,5,6])
	then begin
		if (not result.ok) and (((is_same=0) and point_on_line(p,points_2_line(t^.p1,t^.p2))) or (is_same=4)) then begin
			new(result.t[1]);
			result.t[1]^.flag:=tf_init;
			result.t[1]^.p1:=t^.p3;
			result.t[1]^.p2:=t^.p1;
			result.t[1]^.p3:=p;
			new(result.t[2]);
			result.t[2]^.flag:=tf_init;
			result.t[2]^.p1:=t^.p3;
			result.t[2]^.p2:=p;
			result.t[2]^.p3:=t^.p2;
			new(result.l);
			result.l^:=points_2_line(t^.p1,t^.p2);
			result.ok:=true;
		end;
		if (not result.ok) and (((is_same=0) and point_on_line(p,points_2_line(t^.p2,t^.p3))) or (is_same=5)) then begin
			new(result.t[1]);
			result.t[1]^.flag:=tf_init;
			result.t[1]^.p1:=t^.p1;
			result.t[1]^.p2:=t^.p2;
			result.t[1]^.p3:=p;
			new(result.t[2]);
			result.t[2]^.flag:=tf_init;
			result.t[2]^.p1:=t^.p1;
			result.t[2]^.p2:=p;
			result.t[2]^.p3:=t^.p3;
			new(result.l);
			result.l^:=points_2_line(t^.p2,t^.p3);
			result.ok:=true;
		end;
		if (not result.ok) and (((is_same=0) and point_on_line(p,points_2_line(t^.p3,t^.p1))) or (is_same=6)) then begin
			new(result.t[1]);
			result.t[1]^.flag:=tf_init;
			result.t[1]^.p1:=t^.p2;
			result.t[1]^.p2:=t^.p3;
			result.t[1]^.p3:=p;
			new(result.t[2]);
			result.t[2]^.flag:=tf_init;
			result.t[2]^.p1:=t^.p2;
			result.t[2]^.p2:=p;
			result.t[2]^.p3:=t^.p1;
			new(result.l);
			result.l^:=points_2_line(t^.p3,t^.p1);
			result.ok:=true;
		end;
		if not lineonly then
  		if (not result.ok) then begin
  			new(result.t[1]);
  			result.t[1]^.flag:=tf_init;
  			result.t[1]^.p1:=t^.p1;
  			result.t[1]^.p2:=t^.p2;
  			result.t[1]^.p3:=p;
  			new(result.t[2]);
  			result.t[2]^.flag:=tf_init;
  			result.t[2]^.p1:=t^.p2;
  			result.t[2]^.p2:=t^.p3;
  			result.t[2]^.p3:=p;
  			new(result.t[3]);
  			result.t[3]^.flag:=tf_init;
  			result.t[3]^.p1:=t^.p3;
  			result.t[3]^.p2:=t^.p1;
  			result.t[3]^.p3:=p;
  			result.ok:=true;
  		end;
	end;
{$ifndef FKT_RESULT}
	split_tri:=result;
{$endif}
end;

function split_tri(t:ptriangle;p:tpoint):tsplit_tri;
begin
  split_tri:=split_tri(t,p,false);
end;

function tri_has_side(t:ptriangle;l:pline):boolean;
{stellt fest, ob die Seite in einem Dreieck enthalten ist}
var d:array[1..2,1..3] of float3d;
begin
	d[1,1]:=vec_len(point_point(l^.p1,t^.p1));
	d[2,1]:=vec_len(point_point(l^.p2,t^.p1));
	d[1,2]:=vec_len(point_point(l^.p1,t^.p2));
	d[2,2]:=vec_len(point_point(l^.p2,t^.p2));
	d[1,3]:=vec_len(point_point(l^.p1,t^.p3));
	d[2,3]:=vec_len(point_point(l^.p2,t^.p3));
	tri_has_side:=(d[1,1]*d[1,2]*d[1,3]=0) and (d[2,1]*d[2,2]*d[2,3]=0);
end;

function split_line(t:ptriangle;l:pline;p:tpoint):tsplit_tri;
{Dreiecksunterteilung durch einen auf einer Seite liegenden Punkt}
var i:byte;
{$ifndef FKT_RESULT}
		result:tsplit_tri;
{$endif}		
begin
	for i:=1 to 3 do
		result.t[i]:=NIL;
  if tri_has_side(t,l) then
  	result:=split_tri(t,p);
{$ifndef FKT_RESULT}
  split_line:=result;
{$endif}
end;

procedure inside_out(t:ptriangle);
{Umkehr der Facettenoberflaeche}
var p:tpoint;
begin
	p:=t^.p3;
	t^.p3:=t^.p2;
	t^.p2:=p;
end;

function create_poly(no_points:poly_idx3d;copy_poly:ppoly):ppoly;
{$ifndef FKT_RESULT}
var result:ppoly;
{$endif}
begin
  getmem(result,no_points*sizeof(tpoint));
  if assigned(copy_poly) then
  	move(copy_poly^,result^,no_points*sizeof(tpoint));
{$ifndef FKT_RESULT}
  create_poly:=result;
{$endif}
end;

function create_poly(no_points:poly_idx3d):ppoly;
begin
  create_poly:=create_poly(no_points,NIL);
end;

function free_poly(poly:ppoly;no_points:poly_idx3d):ppoly;
begin
	freemem(poly,no_points*sizeof(tpoint));
	free_poly:=Nil;
end;

procedure insert_poly_point(var no_points:poly_idx3d; var poly:ppoly; p:tpoint);
var idx:poly_idx3d;
		new_poly:ppoly;
		inserted:boolean;
begin
  inserted:=false;
  for idx:=2 to no_points do
  	if not inserted then
  	  if (not (same_point(p,poly^[idx-2]) or same_point(p,poly^[idx-1]))) and
  	  	point_on_line(p,points_2_line(poly^[idx-2],poly^[idx-1]))
  	  then begin
  	  	new_poly:=create_poly(no_points+1);
  	  	move(poly^[0],new_poly^[0],(idx-1)*sizeof(tpoint));
  	  	new_poly^[idx-1]:=p;
  	  	move(poly^[idx-1],new_poly^[idx],(no_points-idx+1)*sizeof(tpoint));
  	  	free_poly(poly,no_points);
  	  	inc(no_points);
  	  	poly:=new_poly;
  	    inserted:=true;
  	  end;
end;

end.
